from __future__ import annotations

import sys
from argparse import Namespace
from asyncio import AbstractEventLoop
from collections.abc import Awaitable, Coroutine, Iterable, Iterator, Sequence
from dataclasses import dataclass
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar, Union, overload

from mpyc import asyncoro, sectypes
from mpyc.finfields import FiniteFieldArray, PrimeFieldElement
from mpyc.sectypes import (
    SecureFiniteField,
    SecureFiniteFieldArray,
    SecureFixedPoint,
    SecureFixedPointArray,
    SecureInteger,
    SecureIntegerArray,
    SecureObject,
)

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

from tno.mpc.mpyc.stubs._sectypes import BaseSecureFloat

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

TypePlaceholder = TypeVar("TypePlaceholder")
SecureObjectT = TypeVar("SecureObjectT", bound=SecureObject)
SecureIntT = TypeVar("SecureIntT", bound=SecureInteger)
SecureFxpT = TypeVar("SecureFxpT", bound=SecureFixedPoint)
PossibleList = Union[TypePlaceholder, list[TypePlaceholder]]

class Runtime:
    def __init__(self, pid: int, parties: Iterable[Party], options: Namespace) -> None:
        self.pid = pid
        self.parties = tuple(parties)
        self.options = options
        self.threshold: int
        self._loop: AbstractEventLoop
        self._pc_level: int
        self._logging_enabled: bool
        self._program_counter: list[int]
        self.aggregate_load: float
        self.start_time: float | None
    coroutine = staticmethod(asyncoro.mpc_coro)
    returnType = staticmethod(asyncoro.returnType)
    SecFld = staticmethod(sectypes.SecFld)
    SecInt = staticmethod(sectypes.SecInt)
    SecFxp = staticmethod(sectypes.SecFxp)

    # Some of the functions below are actually async functions. However, MPyC
    # does not return Awaitables from these functions but rather SecureObjects
    # with an Awaitable `value` field. This causes issues with type hinting.
    # It seems that this workaround allows MyPy to work as intended.
    async def barrier(self, name: str | None = ...) -> None: ...
    async def __aenter__(self) -> Awaitable[Self]: ...
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> Awaitable[bool | None]: ...
    async def transfer(
        self,
        obj: Any,
        senders: Sequence[int] | int | None = ...,
        receivers: Sequence[int] | int | None = ...,
        sender_receivers: dict[int, int] | tuple[int, int] | None = ...,
    ) -> Any: ...
    @overload
    def input(
        self,
        x: SecureObjectT,
        senders: int,
    ) -> SecureObjectT: ...
    @overload
    def input(
        self,
        x: SecureObjectT,
        senders: Sequence[int] | None = ...,
    ) -> list[SecureObjectT]: ...
    @overload
    def input(
        self,
        x: Sequence[SecureObjectT],
        senders: int,
    ) -> list[SecureObjectT]: ...
    @overload
    def input(
        self,
        x: Sequence[SecureObjectT],
        senders: Sequence[int] | None = ...,
    ) -> list[list[SecureObjectT]]: ...
    @overload
    def _random(
        self, sftype: type[SecureObjectT], bound: int | None
    ) -> SecureObjectT: ...
    @overload
    def _random(
        self, sftype: type[PrimeFieldElement], bound: int | None
    ) -> PrimeFieldElement: ...
    @overload
    def _randoms(
        self, sftype: type[SecureObjectT], n: int, bound: int | None = ...
    ) -> list[SecureObjectT]: ...
    @overload
    def _randoms(
        self, sftype: type[PrimeFieldElement], n: int, bound: int | None = ...
    ) -> list[PrimeFieldElement]: ...
    @overload
    def random_bits(
        self, sftype: type[SecureObjectT], n: int, signed: bool = ...
    ) -> list[SecureObjectT]: ...
    @overload
    def random_bits(
        self, sftype: type[PrimeFieldElement], n: int, signed: bool = ...
    ) -> Awaitable[list[PrimeFieldElement]]: ...
    def unit_vector(self, a: SecureObjectT, n: int) -> list[SecureObjectT]: ...
    @overload
    def matrix_prod(
        self,
        A: Sequence[Sequence[SecureObjectT]],
        B: Sequence[Sequence[float]],
        tr: bool = ...,
    ) -> list[list[SecureObjectT]]: ...
    @overload
    def matrix_prod(
        self,
        A: Sequence[Sequence[float]],
        B: Sequence[Sequence[SecureObjectT]],
        tr: bool = ...,
    ) -> list[list[SecureObjectT]]: ...
    @overload
    def matrix_prod(
        self,
        A: Sequence[Sequence[SecureObjectT]],
        B: Sequence[Sequence[SecureObjectT]],
        tr: bool = ...,
    ) -> list[list[SecureObjectT]]: ...
    async def shutdown(self) -> None: ...
    # region output overloads
    # single element as input
    @overload
    async def output(
        self,
        obj: PrimeFieldElement | SecureFiniteField,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: bool = False,
    ) -> PrimeFieldElement: ...
    @overload
    async def output(
        self,
        obj: SecureInteger | SecureFixedPoint,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        *,
        raw: Literal[True],
    ) -> PrimeFieldElement: ...
    @overload
    async def output(
        self,
        obj: SecureInteger,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: Literal[False] = ...,
    ) -> int: ...
    @overload
    async def output(
        self,
        obj: SecureInteger,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: bool = False,
    ) -> int | PrimeFieldElement: ...
    @overload
    async def output(
        self,
        obj: SecureFixedPoint,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: Literal[False] = ...,
    ) -> float: ...
    @overload
    async def output(
        self,
        obj: SecureFixedPoint,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: bool = False,
    ) -> float | PrimeFieldElement: ...
    @overload
    async def output(
        self,
        obj: BaseSecureFloat,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: bool = False,
    ) -> float: ...

    # lists section
    @overload
    async def output(
        self,
        obj: list[PrimeFieldElement] | list[SecureFiniteField],
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: bool = False,
    ) -> list[PrimeFieldElement]: ...
    @overload
    async def output(
        self,
        obj: list[SecureInteger] | list[SecureFixedPoint],
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        *,
        raw: Literal[True],
    ) -> list[PrimeFieldElement]: ...
    @overload
    async def output(
        self,
        obj: list[SecureInteger],
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: Literal[False] = ...,
    ) -> list[int]: ...
    @overload
    async def output(
        self,
        obj: list[SecureInteger],
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: bool = False,
    ) -> list[int] | list[PrimeFieldElement]: ...
    @overload
    async def output(
        self,
        obj: list[SecureFixedPoint],
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: Literal[False] = ...,
    ) -> list[float]: ...
    @overload
    async def output(
        self,
        obj: list[SecureFixedPoint],
        receivers: None | int | list[int],
        threshold: None | int = ...,
        raw: bool = False,
    ) -> list[float] | list[PrimeFieldElement]: ...
    @overload
    async def output(
        self,
        obj: Sequence[BaseSecureFloat],
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: bool = False,
    ) -> list[float]: ...

    # numpy section
    @overload
    async def output(
        self,
        obj: SecureFiniteFieldArray,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: bool = False,
    ) -> FiniteFieldArray: ...
    @overload
    async def output(
        self,
        obj: SecureIntegerArray | SecureFixedPointArray,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: Literal[True] = ...,
    ) -> FiniteFieldArray: ...
    @overload
    async def output(
        self,
        obj: SecureIntegerArray | SecureFixedPointArray,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: Literal[False] = ...,
    ) -> NDArray[np.object_]: ...
    @overload
    async def output(
        self,
        obj: SecureIntegerArray | SecureFixedPointArray,
        receivers: None | int | list[int] = ...,
        threshold: None | int = ...,
        raw: bool = False,
    ) -> NDArray[np.object_] | FiniteFieldArray: ...

    # endregion

    @overload
    def _reshare(self, x: SecureObjectT) -> SecureObjectT: ...
    @overload
    def _reshare(self, x: list[SecureObjectT]) -> list[SecureObjectT]: ...
    @overload
    def _reshare(
        self, x: list[PrimeFieldElement]
    ) -> Awaitable[list[PrimeFieldElement]]: ...
    @overload
    def _reshare(self, x: PrimeFieldElement) -> Awaitable[PrimeFieldElement]: ...
    @overload
    def convert(
        self, x: Sequence[SecureObjectT], t_type: type[TypePlaceholder]
    ) -> list[TypePlaceholder]: ...
    @overload
    def convert(
        self, x: SecureObjectT, t_type: type[TypePlaceholder]
    ) -> TypePlaceholder: ...
    @overload
    def trunc(
        self, x: list[SecureObjectT], f: int | None = ..., l: int | None = ...
    ) -> list[SecureObjectT]: ...
    @overload
    def trunc(
        self, x: SecureObjectT, f: int | None = ..., l: int | None = ...
    ) -> SecureObjectT: ...
    @overload
    def trunc(
        self, x: PrimeFieldElement, f: int | None = ..., l: int | None = ...
    ) -> Awaitable[PrimeFieldElement]: ...
    @overload
    def trunc(
        self, x: list[PrimeFieldElement], f: int | None = ..., l: int | None = ...
    ) -> Awaitable[list[PrimeFieldElement]]: ...
    async def is_zero_public(self, a: SecureObjectT) -> bool: ...
    def is_zero(self, a: SecureObjectT) -> SecureObjectT: ...
    def sub(self, a: SecureObjectT, b: SecureObjectT | float) -> SecureObjectT: ...
    def add(self, a: SecureObjectT, b: SecureObjectT | float) -> SecureObjectT: ...
    def reciprocal(self, a: SecureObjectT) -> SecureObjectT: ...
    def lsb(self, a: SecureObjectT) -> SecureObjectT: ...
    def mod(self, a: SecureObjectT, b: SecureObjectT) -> SecureObjectT: ...
    @overload
    def if_else(
        self, c: SecureObjectT, x: SecureObjectT, y: SecureObjectT
    ) -> SecureObjectT: ...
    @overload
    def if_else(
        self, c: SecureObjectT, x: list[SecureObjectT], y: list[SecureObjectT]
    ) -> list[SecureObjectT]: ...
    def _if_else_list(
        self, c: SecureObjectT, x: list[SecureObjectT], y: list[SecureObjectT]
    ) -> list[SecureObjectT]: ...
    def _convert(
        self, x: SecureObject, ttype: type[SecureObjectT]
    ) -> SecureObjectT: ...
    def lt(self, a: SecureObjectT, b: SecureObjectT | float) -> SecureObjectT: ...
    def eq(self, a: SecureObjectT, b: SecureObjectT | float) -> SecureObjectT: ...
    def ge(self, a: SecureObjectT, b: SecureObjectT | float) -> SecureObjectT: ...
    def abs(self, a: SecureObjectT, l: int | None = ...) -> SecureObjectT: ...
    def sgn(
        self, a: SecureObjectT, l: int = ..., LT: bool = ..., EQ: bool = ...
    ) -> SecureObjectT: ...
    def to_bits(self, a: SecureObjectT, l: int | None) -> list[SecureObjectT]: ...
    def from_bits(self, x: Sequence[SecureObjectT]) -> SecureObjectT: ...
    @overload
    def max(
        self,
        *x: SecureObjectT,
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> SecureObjectT: ...
    @overload
    def max(  # type: ignore[overload-overlap] # incompatible overload indeed, but the truth
        self,
        x: list[SecureObjectT],
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> SecureObjectT: ...
    @overload
    def max(
        self,
        x: Iterator[SecureObjectT],
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> SecureObjectT: ...
    @overload
    def max(
        self,
        *x: list[SecureObjectT],
        key: None | (
            Callable[[list[SecureObjectT]], SecureObjectT | int | float]
        ) = ...,
    ) -> list[SecureObjectT]: ...
    @overload
    def min(
        self,
        *x: SecureObjectT,
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> SecureObjectT: ...
    @overload
    def min(  # type: ignore[overload-overlap] # incompatible overload indeed, but the truth
        self,
        x: list[SecureObjectT],
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> SecureObjectT: ...
    @overload
    def min(
        self,
        x: Iterator[SecureObjectT],
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> SecureObjectT: ...
    @overload
    def min(
        self,
        *x: list[SecureObjectT],
        key: None | (
            Callable[[list[SecureObjectT]], SecureObjectT | int | float]
        ) = ...,
    ) -> list[SecureObjectT]: ...
    @overload
    def argmax(
        self,
        *x: SecureObjectT,
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> tuple[SecureObjectT, SecureObjectT]: ...
    @overload
    def argmax(  # type: ignore[overload-overlap] # incompatible overload indeed, but the truth
        self,
        x: list[SecureObjectT],
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> tuple[SecureObjectT, SecureObjectT]: ...
    @overload
    def argmax(
        self,
        x: Iterator[SecureObjectT],
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> tuple[SecureObjectT, SecureObjectT]: ...
    @overload
    def argmax(
        self,
        *x: list[SecureObjectT],
        key: None | (
            Callable[[list[SecureObjectT]], SecureObjectT | int | float]
        ) = ...,
    ) -> tuple[SecureObjectT, list[SecureObjectT]]: ...
    @overload
    def argmin(
        self,
        *x: SecureObjectT,
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> tuple[SecureObjectT, SecureObjectT]: ...
    @overload
    def argmin(  # type: ignore[overload-overlap] # incompatible overload indeed, but the truth
        self,
        x: list[SecureObjectT],
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> tuple[SecureObjectT, SecureObjectT]: ...
    @overload
    def argmin(
        self,
        x: Iterator[SecureObjectT],
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
    ) -> tuple[SecureObjectT, SecureObjectT]: ...
    @overload
    def argmin(
        self,
        *x: list[SecureObjectT],
        key: None | (
            Callable[[list[SecureObjectT]], SecureObjectT | int | float]
        ) = ...,
    ) -> tuple[SecureObjectT, list[SecureObjectT]]: ...
    def div(
        self,
        a: SecureObjectT | float,
        b: SecureObjectT | float,
    ) -> SecureObjectT: ...
    def sum(self, x: Iterable[SecureObjectT], start: int = ...) -> SecureObjectT: ...
    def in_prod(
        self, x: Sequence[SecureObjectT], y: Sequence[SecureObjectT]
    ) -> SecureObjectT: ...
    @overload
    def prod(
        self,
        x: Sequence[SecureIntT],
        start: int | PrimeFieldElement | SecureIntT = ...,
    ) -> SecureIntT: ...
    @overload
    def prod(
        self,
        x: Sequence[SecureFxpT],
        start: int | float | PrimeFieldElement | SecureFxpT = ...,
    ) -> SecureFxpT: ...
    @overload
    def prod(
        self,
        x: Sequence[SecureFiniteField],
        start: int | SecureFiniteField = ...,
    ) -> SecureFiniteField: ...
    @overload
    def prod(
        self,
        x: Sequence[PrimeFieldElement],
        start: int | PrimeFieldElement = ...,
    ) -> Awaitable[PrimeFieldElement]: ...
    def all(self, x: Sequence[SecureObjectT]) -> SecureObjectT: ...
    def vector_add(
        self, x: Sequence[SecureObjectT], y: Sequence[SecureObjectT]
    ) -> list[SecureObjectT]: ...
    def vector_sub(
        self, x: Sequence[SecureObjectT], y: Sequence[SecureObjectT]
    ) -> list[SecureObjectT]: ...
    def scalar_mul(
        self,
        a: float | SecureObjectT,
        x: Sequence[SecureObjectT],
    ) -> list[SecureObjectT]: ...
    def mul(self, a: SecureObjectT, b: SecureObjectT) -> SecureObjectT: ...
    @overload
    def schur_prod(
        self, x: Sequence[SecureObjectT], y: Sequence[SecureObjectT]
    ) -> list[SecureObjectT]: ...
    @overload
    def schur_prod(
        self, x: Sequence[PrimeFieldElement], y: Sequence[PrimeFieldElement]
    ) -> Awaitable[list[PrimeFieldElement]]: ...
    def run(self, f: Coroutine[Any, None, TypePlaceholder]) -> TypePlaceholder: ...
    @overload
    def sorted(
        self,
        x: Iterable[SecureObjectT],
        key: None | (Callable[[SecureObjectT], SecureObjectT | int | float]) = ...,
        reverse: bool = ...,
    ) -> list[SecureObjectT]: ...
    @overload
    def sorted(
        self,
        x: Iterable[list[SecureObjectT]],
        key: None | (
            Callable[[list[SecureObjectT]], SecureObjectT | int | float]
        ) = ...,
        reverse: bool = ...,
    ) -> list[list[SecureObjectT]]: ...
    def matrix_sub(
        self,
        A: Sequence[Sequence[SecureObjectT]],
        B: Sequence[Sequence[SecureObjectT]],
        tr: bool = ...,
    ) -> list[list[SecureObjectT]]: ...
    async def start(self) -> None: ...
    def _send_message(self, peer_pid: int, data: Any) -> None: ...
    def _receive_message(self, peer_pid: int) -> Any: ...

    # region Extensive Gather Types (for number of parameters 1, 2, 3 and 4 and nesting depth 7)
    # The type hints in this region have been automatically generated using the `generate_gather_stubs.py` script

    @overload
    def gather(
        self,
        obj_1: SecureObject,
    ) -> Awaitable[PrimeFieldElement]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
    ) -> Awaitable[list[PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[list[list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[list[list[list[PrimeFieldElement]]]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
    ) -> Awaitable[PrimeFieldElement]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
    ) -> Awaitable[list[PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[list[list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[list[list[list[PrimeFieldElement]]]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
    ) -> Awaitable[tuple[PrimeFieldElement, list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[tuple[PrimeFieldElement, list[list[PrimeFieldElement]]]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]]]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
    ) -> Awaitable[tuple[PrimeFieldElement, list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[tuple[PrimeFieldElement, list[list[PrimeFieldElement]]]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
    ) -> Awaitable[tuple[list[PrimeFieldElement], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
    ) -> Awaitable[tuple[list[PrimeFieldElement], list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
    ) -> Awaitable[tuple[list[PrimeFieldElement], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
    ) -> Awaitable[tuple[list[PrimeFieldElement], list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
    ) -> Awaitable[tuple[list[list[PrimeFieldElement]], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
    ) -> Awaitable[tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
    ) -> Awaitable[tuple[list[list[PrimeFieldElement]], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
    ) -> Awaitable[tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
    ) -> Awaitable[tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
    ) -> Awaitable[tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
    ) -> Awaitable[tuple[PrimeFieldElement, list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[tuple[PrimeFieldElement, list[list[PrimeFieldElement]]]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]]]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
    ) -> Awaitable[tuple[PrimeFieldElement, list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[tuple[PrimeFieldElement, list[list[PrimeFieldElement]]]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
    ) -> Awaitable[tuple[list[PrimeFieldElement], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
    ) -> Awaitable[tuple[list[PrimeFieldElement], list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
    ) -> Awaitable[tuple[list[PrimeFieldElement], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
    ) -> Awaitable[tuple[list[PrimeFieldElement], list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
    ) -> Awaitable[tuple[list[list[PrimeFieldElement]], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
    ) -> Awaitable[tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
    ) -> Awaitable[tuple[list[list[PrimeFieldElement]], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
    ) -> Awaitable[tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement]]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
    ) -> Awaitable[tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
    ) -> Awaitable[tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: SecureObject,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: SecureObject,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[tuple[PrimeFieldElement, PrimeFieldElement, PrimeFieldElement]]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, PrimeFieldElement, list[list[list[PrimeFieldElement]]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[PrimeFieldElement], list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[PrimeFieldElement]], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[PrimeFieldElement, list[list[list[PrimeFieldElement]]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], PrimeFieldElement, list[list[PrimeFieldElement]]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[PrimeFieldElement], list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[PrimeFieldElement], list[list[PrimeFieldElement]], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], PrimeFieldElement, list[PrimeFieldElement]]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[PrimeFieldElement]], list[PrimeFieldElement], PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[list[list[list[PrimeFieldElement]]], PrimeFieldElement, PrimeFieldElement]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: SecureObject,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[SecureObject],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[SecureObject]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[SecureObject]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement, PrimeFieldElement, PrimeFieldElement, PrimeFieldElement
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: PrimeFieldElement,
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[PrimeFieldElement],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[PrimeFieldElement]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: SecureObject,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[SecureObject],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[SecureObject]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[SecureObject]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: PrimeFieldElement,
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[PrimeFieldElement],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[PrimeFieldElement]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: SecureObject,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[SecureObject],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[SecureObject]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[SecureObject]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: PrimeFieldElement,
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[PrimeFieldElement],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[PrimeFieldElement]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: SecureObject,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[SecureObject],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[SecureObject]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[SecureObject]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: PrimeFieldElement,
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            PrimeFieldElement,
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[PrimeFieldElement],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[PrimeFieldElement],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[PrimeFieldElement]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[PrimeFieldElement]],
        ]
    ]: ...
    @overload
    def gather(
        self,
        obj_1: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_2: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_3: Sequence[Sequence[Sequence[PrimeFieldElement]]],
        obj_4: Sequence[Sequence[Sequence[PrimeFieldElement]]],
    ) -> Awaitable[
        tuple[
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
            list[list[list[PrimeFieldElement]]],
        ]
    ]: ...
    # endregion

@dataclass
class Party:
    pid: int
    host: str | None = None
    port: int | None = None
    protocol: None = None

mpc: Runtime
